searchData={"items":[{"type":"module","title":"Tarearbol","doc":"`Tarearbol` module provides an interface to run tasks in easy way.","ref":"Tarearbol.html"},{"type":"module","title":"Examples - Tarearbol","doc":"iex> result = Tarearbol.ensure(fn -> raise \"¡?\" end, attempts: 1, raise: false)\n    iex> {:error, %{job: _job, outcome: outcome}} = result\n    iex> {error, _stacktrace} = outcome\n    iex> error\n    %RuntimeError{message: \"¡?\"}","ref":"Tarearbol.html#module-examples"},{"type":"function","title":"Tarearbol.drain/1","doc":"Executes all the scheduled tasks immediately, cleaning up the queue.","ref":"Tarearbol.html#drain/1"},{"type":"function","title":"Tarearbol.ensure/2","doc":"Ensures the task to be completed; restarts it when necessary.\n\nPossible options:\n- `attempts` [_default:_ `:infinity`] Might be any of `@Tarearbol.Utils.interval`\n  type (`5` for five attempts, `:random` for the random amount etc)\n- `delay` [_default:_ `1 msec`]. Might be any of `@Tarearbol.Utils.interval`\n  type (`1_000` or `1.0` for one second, `:timeout` for five seconds etc)\n- `on_success` [_default:_ `nil`], the function to be called on successful\n  execution (`arity ∈ [0, 1]` or tuple `{Mod, fun}` where `fun` is of arity\n  zero or one.) When the arity of given function is `1`, the result of\n  task execution is passed\n- `on_retry` [_default:_ `nil`], same as above, called on retries after\n  insuccessful attempts **or** one of `[:debug, :info, :warn, :error]` atoms\n  to log a retry with default logger\n- `on_fail` [_default:_ `nil`], same as above, called when the task finally\n  failed after `attempts` amount of insuccessful attempts","ref":"Tarearbol.html#ensure/2"},{"type":"function","title":"Tarearbol.ensure!/2","doc":"Same as `Tarearbol.ensure/2`, but it raises on fail and returns the result\n  itself on successful execution.","ref":"Tarearbol.html#ensure!/2"},{"type":"function","title":"Tarearbol.ensure_all/2","doc":"Executes `Tarearbol.ensure_all_streamed/2` and collects tasks results.","ref":"Tarearbol.html#ensure_all/2"},{"type":"function","title":"Tarearbol.ensure_all_streamed/2","doc":"Wrapper for [`Task.Supervisor.async_stream/4`](https://hexdocs.pm/elixir/Task.Supervisor.html#async_stream/4).","ref":"Tarearbol.html#ensure_all_streamed/2"},{"type":"function","title":"Tarearbol.run_at/3","doc":"Runs a task specified by the first argument at a given time.\n\nIf the second parameter is a [`DateTime`] struct, the task will be run once.\nIf the second parameter is a [`Time`] struct, the task will be run at that time\n  on daily basis.","ref":"Tarearbol.html#run_at/3"},{"type":"function","title":"Tarearbol.run_in/3","doc":"Runs a task specified by the first argument in a given interval.\n\nSee [`Tarearbol.ensure/2`] for all possible variants of the `interval` argument.","ref":"Tarearbol.html#run_in/3"},{"type":"function","title":"Tarearbol.spawn/2","doc":"Spawns the task for the immediate async execution.","ref":"Tarearbol.html#spawn/2"},{"type":"function","title":"Tarearbol.spawn_ensured/2","doc":"Spawns an ensured job asynchronously, passing all options given.","ref":"Tarearbol.html#spawn_ensured/2"},{"type":"behaviour","title":"Tarearbol.DynamicManager","doc":"The scaffold implementation to dynamically manage many similar tasks running\nas processes.\n\nIt creates a main supervisor, managing the `GenServer` holding the state and\n`DynamicSupervisor` handling chidren. It has a strategy `:rest_for_one`,\nassuming that if the process holding the state crashes, the children will be\nrestarted.\n\nTypically one calls `use Tarearbol.DynamicManager` and implements at least\n`children_specs/0` callback and receives back supervised tree with a state\nand many processes controlled by `DynamicSupervisor`.\n\nTo see how it works you might try\n\n    defmodule DynamicManager do\n      use Tarearbol.DynamicManager\n\n      def children_specs do\n        for i <- 1..10, do: {\"foo_#{i}\", []}, into: %{}\n      end\n    end\n\n    {:ok, pid} = DynamicManager.start_link()\n\nThe above would spawn `10` children with IDs `\"foo_1\"..\"foo_10\"`.","ref":"Tarearbol.DynamicManager.html"},{"type":"behaviour","title":"Workers Management - Tarearbol.DynamicManager","doc":"`DynamicManager` allows dynamic workers management. It exports three functions\n\n    @spec get(id :: id()) :: Enum.t()\n    @spec put(id :: id(), opts :: Enum.t()) :: pid()\n    @spec del(id :: id()) :: :ok\n    @spec restart(id :: id()) :: :ok\n\nThe semantics of `put/2` arguments is the same as a single `child_spec`,\n`del/1` and `get/1` receive the unique ID of the child and shutdown it or\nreturn it’s payload respectively.","ref":"Tarearbol.DynamicManager.html#module-workers-management"},{"type":"behaviour","title":"Workers Callbacks - Tarearbol.DynamicManager","doc":"Workers are allowed to implement several callbacks to be used to pass messages\n  to them.\n\n- **`peform/2`** is called periodically by the library internals; the interval\n  is set upon worker initialization via `children_specs/1` (static) or `put/2`\n  (dynamic); the interval set to `0` suppresses periodic invocations\n- **`call/3`** to handle synchronous message send to worker\n- **`cast/2`** to handle asynchronous message send to worker\n- **`terminate/2`** to handle worker process termination\n\nAll the above should return a value of `t:Tarearbol.DynamicManager.response/0` type.\n\nAlso, the implementing module might use a custom initialization function\n  to e. g. dynamically build payload. Is should be passed to `use DynamicManager`\n  as a parameter `init: handler` and might be a tuple `{module(), function(), arity()}` or\n  a captured function `&MyMod.my_init/1`. Arities 0, 1 and 2 are allowed, as described by\n  `t:Tarearbol.DynamicManager.init_handler/0` type.\n\nThe worker process will call this function from `c:GenServer.handle_continue/2` callback.","ref":"Tarearbol.DynamicManager.html#module-workers-callbacks"},{"type":"callback","title":"Tarearbol.DynamicManager.call/3","doc":"The method to implement to support explicit `GenServer.call/3` on the wrapping worker.","ref":"Tarearbol.DynamicManager.html#c:call/3"},{"type":"callback","title":"Tarearbol.DynamicManager.cast/2","doc":"The method to implement to support explicit `GenServer.cast/2` on the wrapping worker.","ref":"Tarearbol.DynamicManager.html#c:cast/2"},{"type":"callback","title":"Tarearbol.DynamicManager.children_specs/0","doc":"This function is called to retrieve the map of children with name as key\nand a workers as the value.\n\nThe value must be an enumerable with keys among:\n- `:payload` passed as second argument to `perform/2`, default `nil`\n- `:timeout` time between iterations of `perform/2`, default `1` second\n- `:lull` threshold to notify latency in performing, default `1.1`\n  (the threshold is `:lull` times the `:timeout`)\n\nThis function should not care about anything save for producing side effects.\n\nIt will be backed by `DynamicSupervisor`. The value it returns will be put\ninto the state under `children` key.","ref":"Tarearbol.DynamicManager.html#c:children_specs/0"},{"type":"callback","title":"Tarearbol.DynamicManager.handle_state_change/1","doc":"Declares an instance-wide callback to report state; if the startup process\ntakes a while, it’d be run in `handle_continue/2` and this function will be\ncalled after it finishes so that the application might start using it.\n\nIf the application is not interested in receiving state updates, e. g. when\nall it needs from runners is a side effect, there is a default implementation\nthat does nothing.","ref":"Tarearbol.DynamicManager.html#c:handle_state_change/1"},{"type":"callback","title":"Tarearbol.DynamicManager.handle_timeout/1","doc":"Declares a callback to report slow process (when the scheduler cannot process\nin a reasonable time).","ref":"Tarearbol.DynamicManager.html#c:handle_timeout/1"},{"type":"callback","title":"Tarearbol.DynamicManager.perform/2","doc":"The main function, doing all the internal job, supervised.\n\nIt will be called with the child `id` as first argument and the\n`payload` option to child spec as second argument (defaulting to `nil`,\ncan also be ignored if not needed).\n\n#","ref":"Tarearbol.DynamicManager.html#c:perform/2"},{"type":"callback","title":"Return values - Tarearbol.DynamicManager.perform/2","doc":"`perform/2` might return\n\n- `:halt` if it wants to be killed\n- `{:ok, result}` to store the last result and reschedule with default timeout\n- `{:replace, payload}` to replace the payload (state) of the current worker with the new one\n- `{:replace, id, payload}` to replace the current worker with the new one\n- `{{:timeout, timeout}, result}` to store the last result and reschedule in given timeout interval\n- or **_deprecated_** anything else will be treated as a result","ref":"Tarearbol.DynamicManager.html#c:perform/2-return-values"},{"type":"callback","title":"Tarearbol.DynamicManager.terminate/2","doc":"The method that will be called before the worker is terminated.","ref":"Tarearbol.DynamicManager.html#c:terminate/2"},{"type":"type","title":"Tarearbol.DynamicManager.id/0","doc":"Identifier of the child process","ref":"Tarearbol.DynamicManager.html#t:id/0"},{"type":"type","title":"Tarearbol.DynamicManager.init_handler/0","doc":"Post-instantion init handler type, that might be passed to `use DynamicManager` vis `init:`","ref":"Tarearbol.DynamicManager.html#t:init_handler/0"},{"type":"type","title":"Tarearbol.DynamicManager.payload/0","doc":"Payload associated with the worker","ref":"Tarearbol.DynamicManager.html#t:payload/0"},{"type":"type","title":"Tarearbol.DynamicManager.response/0","doc":"Expected response from the `DymanicManager` implementation","ref":"Tarearbol.DynamicManager.html#t:response/0"},{"type":"module","title":"Tarearbol.DynamicManager.Child","doc":"The internal representation of a child process under `DynamicManager` supervision","ref":"Tarearbol.DynamicManager.Child.html"},{"type":"type","title":"Tarearbol.DynamicManager.Child.t/0","doc":"","ref":"Tarearbol.DynamicManager.Child.html#t:t/0"},{"type":"module","title":"Tarearbol.Pool","doc":"The pool of workers built on top of `Tarearbol.DynamicManager`.","ref":"Tarearbol.Pool.html"},{"type":"module","title":"Configuration options: - Tarearbol.Pool","doc":"- **`pool_size`** the size of the pool, default: `5`\n- **`payload`** the default payload for all the workers\n  (for more complex initialization, use `init` option below)\n- **`init`** custom `init` step as described in  `Tarearbol.DynamicManager` docs\n\nThis module exports `defsynch/2` and `defasynch/2` macros allowing to to declare\n  functions that would be managed by a pool behind. Basically, the below would be translated\n  into a message passed to the free worker of the pool.\n\n    defsynch synch(n) do\n      {:ok, payload!() + n}\n    end\n\nBoth macros have three predefined functions declared inside a block\n\n- **`id!`** returning the `id` of the worker invoked\n- **`payload!`** returning the `payload` of the worker invoked\n- **`state!`** returning the `state` of the worker invoked as a tuple `{id, payload}`","ref":"Tarearbol.Pool.html#module-configuration-options"},{"type":"macro","title":"Tarearbol.Pool.defasynch/2","doc":"Declares an asynchronous function with the same name and block, that will be under the hood\n  routed to the free worker to execute. If there is no free worker at the moment,\n  returns `:error` otherwise returns `{:ok, result}` tuple.\n\nThis function might return any response recognizable by `t:Tarearbol.DynamicManager.response/0`.","ref":"Tarearbol.Pool.html#defasynch/2"},{"type":"macro","title":"Tarearbol.Pool.defsynch/2","doc":"Declares a synchronous function with the same name and block, that will be under the hood\n  routed to the free worker to execute. If there is no free worker at the moment,\n  returns `:error` otherwise returns `{:ok, result}` tuple.\n\nThis function might return any response recognizable by `t:Tarearbol.DynamicManager.response/0`.","ref":"Tarearbol.Pool.html#defsynch/2"},{"type":"module","title":"Tarearbol.Calendar","doc":"Handy functions to work with dates.","ref":"Tarearbol.Calendar.html"},{"type":"function","title":"Tarearbol.Calendar.beginning_of/3","doc":"Returns a `DateTime` instance of the beginning of the period given as third\nparameters, related to the `DateTime` instance given as the first parameter\n(defaulted to `DateTime.utc_now/0`).\n\nThe number of periods to shift might be given as the second parameter\n(defaults to `0`.)\n\n_Examples:_\n\n    iex> dt = DateTime.from_unix!(1567091960)\n    ~U[2019-08-29 15:19:20Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, :year)\n    ~U[2019-01-01 00:00:00Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, -2, :year)\n    ~U[2017-01-01 00:00:00Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, :month)\n    ~U[2019-08-01 00:00:00Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, 6, :month)\n    ~U[2020-02-01 00:00:00Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, 4, :day)\n    ~U[2019-09-02 00:00:00Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, :week)\n    ~U[2019-08-26 00:00:00Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, -34, :week)\n    ~U[2018-12-31 00:00:00Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, :hour)\n    ~U[2019-08-29 15:00:00Z]\n    iex> Tarearbol.Calendar.beginning_of(dt, -28*24-16, :hour)\n    ~U[2019-07-31 23:00:00Z]\n    iex> Tarearbol.Calendar.end_of(dt, :hour)\n    ~U[2019-08-29 15:59:59Z]\n    iex> Tarearbol.Calendar.end_of(dt, 5, :month)\n    ~U[2020-01-31 23:59:59Z]\n    iex> Tarearbol.Calendar.end_of(dt, -1, :year)\n    ~U[2018-12-31 23:59:59Z]","ref":"Tarearbol.Calendar.html#beginning_of/3"},{"type":"function","title":"Tarearbol.Calendar.end_of/3","doc":"","ref":"Tarearbol.Calendar.html#end_of/3"},{"type":"module","title":"Tarearbol.Crontab","doc":"Helper functions to work with `cron` syntax.","ref":"Tarearbol.Crontab.html"},{"type":"function","title":"Tarearbol.Crontab.formula/1","doc":"Produces the single formula out of cron record. Might be useful\nfor some external check that requires the single validation call.\n\n_Examples_\n\n    iex> formula = Tarearbol.Crontab.formula(\"42 3 28 08 *\").formula\n    ...> formula |> String.split(\" && \") |> Enum.sort()\n    [\"(day == 28)\", \"(hour == 3)\", \"(minute == 42)\", \"(month == 8)\", \"(rem(day_of_week, 1) == 0)\"]\n\n    iex> Tarearbol.Crontab.formula(\"423 * * * *\")\n    {:error, [minute: {:could_not_parse_field, [\"423\"]}]}","ref":"Tarearbol.Crontab.html#formula/1"},{"type":"function","title":"Tarearbol.Crontab.next/3","doc":"Returns the next `DateTime` the respective `cron` record points to\nwith a precision given as the third argument (default: `:second`.)\n\nIf the first parameter is not given, it assumes _the next after now_.\n\n_Examples_\n\n    iex> dt = DateTime.from_unix!(1567091960)\n    ~U[2019-08-29 15:19:20Z]\n    iex> Tarearbol.Crontab.next(dt, \"42 3 28 08 *\")\n    [\n      origin: ~U[2019-08-29 15:19:20Z],\n      next: ~U[2020-08-28 03:42:00Z],\n      second: 31494160\n    ]\n\nwhere `origin` contains the timestamp to lookup the `next` for, `next`\nis the `DateTime` instance of the next event and `second` is the\n{`precision`, `difference_in_that_precision`}.","ref":"Tarearbol.Crontab.html#next/3"},{"type":"function","title":"Tarearbol.Crontab.next_as_list/3","doc":"Returns the _list_ of all the events after `dt` (default: `DateTime.utc_now/0`.)\n\nThis function calculates the outcome greedily and, while it might be slightly\nfaster than `Tarearbol.Crontab.next_as_stream/3`, it should not be used for\nfrequently recurring cron records (like `\"* * * * *\"`.)","ref":"Tarearbol.Crontab.html#next_as_list/3"},{"type":"function","title":"Tarearbol.Crontab.next_as_stream/3","doc":"Returns the _stream_ of all the events after `dt` (default: `DateTime.utc_now/0`.)\n\nThis function calculates the outcome lazily, returning a stream.\n\nSee `Tarearbol.Crontab.next_as_list/3` for greedy evaluation.","ref":"Tarearbol.Crontab.html#next_as_stream/3"},{"type":"function","title":"Tarearbol.Crontab.parse/1","doc":"Parses the cron string into human-readable representation.\n\n**This function is exported for debugging purposes only, normally one would call `prepare/1` instead.**\n\nInput format: [\"minute hour day/month month day/week\"](https://crontab.guru/).\n\n_Examples:_\n\n    iex> Tarearbol.Crontab.parse \"10-30/5 */4 1 */1 6,7\"\n    %Tarearbol.Crontab{\n      day: \"(day == 1)\",\n      day_of_week: \"(day_of_week == 6 || day_of_week == 7)\",\n      hour: \"(rem(hour, 4) == 0)\",\n      minute: \"(rem(minute, 5) == 0 && minute >= 10 && minute <= 30)\",\n      month: \"(rem(month, 1) == 0)\"\n    }\n\n_In case of malformed input:_\n\n    iex> Tarearbol.Crontab.parse \"10-30/5 */4 1 */1 6d,7\"\n    %Tarearbol.Crontab{\n      day: \"(day == 1)\",\n      day_of_week: {:error, {:could_not_parse_integer, \"6d\"}},\n      hour: \"(rem(hour, 4) == 0)\",\n      minute: \"(rem(minute, 5) == 0 && minute >= 10 && minute <= 30)\",\n      month: \"(rem(month, 1) == 0)\"\n    }","ref":"Tarearbol.Crontab.html#parse/1"},{"type":"function","title":"Tarearbol.Crontab.prepare/1","doc":"Parses the cron string into `Tarearbol.Crontab.t()` struct.\n\nInput format: [\"minute hour day/month month day/week\"](https://crontab.guru/).","ref":"Tarearbol.Crontab.html#prepare/1"},{"type":"function","title":"Tarearbol.Crontab.to_cron/1","doc":"Converts the `Time` instance into daily-execution cron string","ref":"Tarearbol.Crontab.html#to_cron/1"},{"type":"type","title":"Tarearbol.Crontab.t/0","doc":"Internal representation of the record in cron file","ref":"Tarearbol.Crontab.html#t:t/0"},{"type":"module","title":"Tarearbol.Scheduler","doc":"Cron-like task scheduler. Accepts both static and dynamic configurations.\n\n#","ref":"Tarearbol.Scheduler.html"},{"type":"module","title":"Usage - Tarearbol.Scheduler","doc":"Add `Tarearbol.Scheduler` to the list of supervised workers. It would attempt\nto read the static configuration (see below) and start the `DynamicSupervisor`\nwith all the scheduled jobs as supervised workers.\n\nThe `runner` is the function of arity zero, that should return `{:ok, result}`\ntuple upon completion. The job will be rescheduled according to its schedule.\n\nThe last result returned will be stored in the state and might be retrieved\nlater with `get/1` passing the job name.\n\n#","ref":"Tarearbol.Scheduler.html#module-usage"},{"type":"module","title":"Static Configuration - Tarearbol.Scheduler","doc":"Upon starts it looks up `:tarearbol` section of `Mix.Project` for\n`:jobs` and `:jobs_file` keys. The latter has a default `.tarearbol.exs`.\nThis won’t work with releases.\n\nAlso it looks up `:tarearbol, :jobs` section of `config.exs`. Everything found\nis unioned. Jobs with the same names are overriden, the file has precedence\nover project config, the application config has least precedence.\n\nIf found, jobs as a list of tuples of `{name, runner, schedule}` are scheduled.\nThese are expected to be in the following form.\n\n- `name` might be whatever, used to refer to the job during it’s lifetime\n- `runner` might be either `{module, function}` tuple or a reference to the function of arity zero (`&Foo.bar/0`)\n- `schedule` in standard cron notation, see https://crontab.guru\n\n#","ref":"Tarearbol.Scheduler.html#module-static-configuration"},{"type":"module","title":"Dynamic Configuration - Tarearbol.Scheduler","doc":"Use `Tarearbol.Scheduler.push/3`, `Tarearbol.Scheduler.pop/1` to add/remove jobs\ntemporarily and/or `Tarearbol.Scheduler.push!/3`, `Tarearbol.Scheduler.pop!/1` to\nreflect changes in the configuration file.\n\n```elixir\nTarearbol.Scheduler.push(TestJob, &Foo.bar/0, \"3-5/1 9-18 * * 6-7\")\n```","ref":"Tarearbol.Scheduler.html#module-dynamic-configuration"},{"type":"function","title":"Tarearbol.Scheduler.active_jobs/0","doc":"","ref":"Tarearbol.Scheduler.html#active_jobs/0"},{"type":"function","title":"Tarearbol.Scheduler.asynch_call/2","doc":"Performs a `GenServer.cast/2` to the worker specified by `id`.\n\n`c:Tarearbol.DynamicManager.cast/2` callback should be implemented for this to work.","ref":"Tarearbol.Scheduler.html#asynch_call/2"},{"type":"function","title":"Tarearbol.Scheduler.child_spec/1","doc":"Returns a specification to start this module under a supervisor.\n\nSee `Supervisor`.","ref":"Tarearbol.Scheduler.html#child_spec/1"},{"type":"function","title":"Tarearbol.Scheduler.del/1","doc":"Dynamically removes a supervised worker implementing `Tarearbol.DynamicManager`\nbehaviour from the list of supervised children\n\nIf `distributed: true` parameter was given to `use Tarearbol.DynamicManager`,\n  deletes the worker from all the nodes managed by `Cloister`. `:cloister` dependency\n  must be added to a project to use this feature.","ref":"Tarearbol.Scheduler.html#del/1"},{"type":"function","title":"Tarearbol.Scheduler.get/1","doc":"Retrieves the information (`payload`, `timeout`, `lull` etc.) assotiated with\nthe supervised worker","ref":"Tarearbol.Scheduler.html#get/1"},{"type":"function","title":"Tarearbol.Scheduler.multidel/1","doc":"Dynamically removes a supervised worker implementing `Tarearbol.DynamicManager`\n  behaviour from the list of supervised children on all the nodes managed by `Cloister`.\n\nUse `distributed: true` parameter in call to `use Tarearbol.DynamicManager`\n  and regular `del/1` instead.","ref":"Tarearbol.Scheduler.html#multidel/1"},{"type":"function","title":"Tarearbol.Scheduler.multiput/2","doc":"Dynamically adds a supervised worker implementing `Tarearbol.DynamicManager`\n  behaviour to the list of supervised children on all the nodes managed by `Cloister`.\n\nUse `distributed: true` parameter in call to `use Tarearbol.DynamicManager`\n  and regular `put/2` instead.","ref":"Tarearbol.Scheduler.html#multiput/2"},{"type":"function","title":"Tarearbol.Scheduler.pop/1","doc":"Removes the scheduled job from the schedule by `id`.\n\nFor the implementation that survives restarts use `pop!/1`.","ref":"Tarearbol.Scheduler.html#pop/1"},{"type":"function","title":"Tarearbol.Scheduler.pop!/1","doc":"Removes the scheduled job from the schedule by `id` and updated the configuration.\n\nFor the implementation that removes jobs temporarily, use `pop!/1`.","ref":"Tarearbol.Scheduler.html#pop!/1"},{"type":"function","title":"Tarearbol.Scheduler.push/3","doc":"Creates and temporarily pushes the job to the list of currently scheduled jobs.\n\nFor the implementation that survives restarts use `push!/3`.","ref":"Tarearbol.Scheduler.html#push/3"},{"type":"function","title":"Tarearbol.Scheduler.push!/3","doc":"Creates and pushes the job to the list of currently scheduled jobs, updates\nthe permanent list of scheduled jobs.\n\nFor the implementation that temporarily pushes a job, use `push/3`.","ref":"Tarearbol.Scheduler.html#push!/3"},{"type":"function","title":"Tarearbol.Scheduler.put/2","doc":"Dynamically adds a supervised worker implementing `Tarearbol.DynamicManager`\n  behaviour to the list of supervised children.\n\nUnlike `put_new/3`, this function would have the child replaced (shut down\nby `id` and started again with options given.)\n\nIf `distributed: true` parameter was given to `use Tarearbol.DynamicManager`,\n  puts the worker into all the nodes managed by `Cloister`. `:cloister` dependency\n  must be added to a project to use this feature.","ref":"Tarearbol.Scheduler.html#put/2"},{"type":"function","title":"Tarearbol.Scheduler.put_new/2","doc":"Dynamically adds a supervised worker implementing `Tarearbol.DynamicManager`\n  behaviour to the list of supervised children if and only if it does not exist yet.\n\nIf `distributed: true` parameter was given to `use Tarearbol.DynamicManager`,\n  puts the worker into all the nodes managed by `Cloister`. `:cloister` dependency\n  must be added to a project to use this feature.","ref":"Tarearbol.Scheduler.html#put_new/2"},{"type":"function","title":"Tarearbol.Scheduler.restart/0","doc":"Restarts the `DynamicManager` to the clean state","ref":"Tarearbol.Scheduler.html#restart/0"},{"type":"function","title":"Tarearbol.Scheduler.start_link/1","doc":"Starts the `DynamicSupervisor` and its helpers to manage dynamic children","ref":"Tarearbol.Scheduler.html#start_link/1"},{"type":"function","title":"Tarearbol.Scheduler.synch_call/2","doc":"Performs a `GenServer.call/3` to the worker specified by `id`.\n\n`c:Tarearbol.DynamicManager.call/3` callback should be implemented for this to work.","ref":"Tarearbol.Scheduler.html#synch_call/2"},{"type":"type","title":"Tarearbol.Scheduler.once_schedule/0","doc":"Type of possible job schedules to be run once: `Time` to be executed once or\namount of milliseconds to execute after","ref":"Tarearbol.Scheduler.html#t:once_schedule/0"},{"type":"type","title":"Tarearbol.Scheduler.repeated_schedule/0","doc":"Type of possible job schedules to be run repeatedly: binary cron format\nor `DateTime` for the daily execution","ref":"Tarearbol.Scheduler.html#t:repeated_schedule/0"},{"type":"type","title":"Tarearbol.Scheduler.runner/0","doc":"Type of the job runner, an `{m, f}` tuple or a function of arity zero,\nreturning one of the outcomes below","ref":"Tarearbol.Scheduler.html#t:runner/0"},{"type":"type","title":"Tarearbol.Scheduler.schedule/0","doc":"Combined type for the all schedules possible","ref":"Tarearbol.Scheduler.html#t:schedule/0"},{"type":"module","title":"Tarearbol.Scheduler.Job","doc":"A struct holding the job description. Used internally by `Tarearbol.Scheduler`\nto preserve a list of scheduled jobs.","ref":"Tarearbol.Scheduler.Job.html"},{"type":"function","title":"Tarearbol.Scheduler.Job.create/3","doc":"Produces a `Tarearbol.Scheduler.Job` by parameters given","ref":"Tarearbol.Scheduler.Job.html#create/3"},{"type":"type","title":"Tarearbol.Scheduler.Job.t/0","doc":"The struct containing the information about the job","ref":"Tarearbol.Scheduler.Job.html#t:t/0"},{"type":"exception","title":"Tarearbol.TaskFailedError","doc":"The generic exception thrown from any function\ndealing with `Task` handling.","ref":"Tarearbol.TaskFailedError.html"},{"type":"extras","title":"Tarearbol","doc":"# Tarearbol\n\n- **Scaffold for managing many children under `Tarearbol.DynamicManager` wrapping `DynamicSupervisor` with a pleasure**\n- **Lightweight task manager, allowing retries, callbacks, assurance that the task succeeded, and more**\n- **Lightweight scalable cron-like job scheduler `Tarearbol.Scheduler`**","ref":"intro.html"},{"type":"extras","title":"Dynamic Workers Management - Tarearbol","doc":"#","ref":"intro.html#dynamic-workers-management"},{"type":"extras","title":"Features - Tarearbol","doc":"- Scaffold for the supervised `DynamicSupervisor`\n- No code for managing processes required\n- Implementation is as easy as 4 callbacks\n- Behaviour-driven consumers\n- Consistent state and callbacks on state changes (like restarting)\n\n### [Dynamic Workers Management Examples](dynamic_workers_management.html)","ref":"intro.html#features"},{"type":"extras","title":"Task Management - Tarearbol","doc":"#","ref":"intro.html#task-management"},{"type":"extras","title":"Features - Tarearbol","doc":"- Task supervision tree for granted\n- Infinite retries until succeeded\n- Async execution of a single task with retries\n- Async execution of many tasks with retries\n- Limited amount of retries\n- Delay between retries\n- Callbacks on `on_retry`, `on_success`, `on_fail`\n- Configurable “success” treatment\n- Task scheduling and draining, cron-like execution\n\n### [Task Management Examples](task_management.html)","ref":"intro.html#features"},{"type":"extras","title":"Job Management - Tarearbol","doc":"#","ref":"intro.html#job-management"},{"type":"extras","title":"Features - Tarearbol","doc":"- Fully supported [cron syntax](https://crontab.guru/)\n- Job callback with an ability to cancel and reschedule jobs\n- Easy configuration via project, config, or external file\n- No persistent storage required\n- `Stream` with all the upcoming cron events by cron record\n- `Tarearbol.Calendar.{beginning_of,end_of}/2` with full `Calendar` support\n\n### [Job Management Examples](cron_management.html)","ref":"intro.html#features"},{"type":"extras","title":"Task Management","doc":"# Task Management","ref":"task_management.html"},{"type":"extras","title":"Task supervision tree for granted - Task Management","doc":"Add `:tarearbol` to the list of applications and you are all set.","ref":"task_management.html#task-supervision-tree-for-granted"},{"type":"extras","title":"Infinite retries - Task Management","doc":"```elixir\nTarearbol.ensure fn ->\n  unless Enum.random(1..100) == 42, do: raise \"Incorrect answer\"\n  {:ok, 42}\nend\n\n# some bad-case logging\n{:ok, 42}\n```","ref":"task_management.html#infinite-retries"},{"type":"extras","title":"Async execution with retries - Task Management","doc":"```elixir\nTarearbol.spawn_ensured({Process, :sleep, [1_000]},\n  attempts: 1,\n  on_success: {IO, :inspect, [[label: \"Result\"]]})\n\n%Task{\n  owner: #PID<0.248.0>,\n  pid: #PID<0.271.0>,\n  ref: #Reference<0.1612468028.2321547270.14001>\n}\n#  \nResult: :ok\n```","ref":"task_management.html#async-execution-with-retries"},{"type":"extras","title":"Async execution of many tasks with retries - Task Management","doc":"```elixir\nres = 1..20\n      |> Enum.map(fn i ->\n        fn -> Process.sleep(Enum.random(1..i)); i end\n      end)\n      |> Tarearbol.ensure_all(attempts: 1)\n\n[{:ok, 1}, {:ok, 2}, ..., {:ok, 20}]\n```\n\n**NB** the above is synchronized against main process.","ref":"task_management.html#async-execution-of-many-tasks-with-retries"},{"type":"extras","title":"Limited amount of retries - Task Management","doc":"```elixir\nTarearbol.ensure fn ->\n  raise \"Incorrect answer\"\nend, attempts: 10\n\n# some bad-case logging\n{:error,\n %{job: #Function<20.87737649/0 in :erl_eval.expr/5>,\n   outcome: {%RuntimeError{message: \"Incorrect answer\"},\n    [{:erl_eval, :do_apply, 6, [file: 'erl_eval.erl', line: 668]},\n     {Task.Supervised, :do_apply, 2,\n      [file: 'lib/task/supervised.ex', line: 85]},\n     {Task.Supervised, :reply, 5, [file: 'lib/task/supervised.ex', line: 36]},\n     {:proc_lib, :init_p_do_apply, 3, [file: 'proc_lib.erl', line: 247]}]}}}\n```","ref":"task_management.html#limited-amount-of-retries"},{"type":"extras","title":"Delay between retries - Task Management","doc":"```elixir\nTarearbol.ensure fn ->\n  unless Enum.random(1..100) == 42, do: raise \"Incorrect answer\"\n  {:ok, 42}\nend, delay: 1000\n\n# some slow bad-case logging\n{:ok, 42}\n```","ref":"task_management.html#delay-between-retries"},{"type":"extras","title":"Callbacks - Task Management","doc":"```elixir\nTarearbol.ensure fn ->\n  unless Enum.random(1..100) == 42, do: raise \"Incorrect answer\"\n  {:ok, 42}\nend, on_success: fn data -> IO.inspect(data, label: \"★\") end,\n     on_retry: {IO, :inspect, [[label: \"☆\"]]} end\n\n# some slow bad-case logging\n# ⇓⇓⇓⇓ one or more of ⇓⇓⇓⇓\n☆: %{cause: :on_raise,\n  data: {%RuntimeError{message: \"Incorrect answer\"},\n   [{:erl_eval, :do_apply, 6, [file: 'erl_eval.erl', line: 670]},\n    {:erl_eval, :exprs, 5, [file: 'erl_eval.erl', line: 122]},\n    {Task.Supervised, :do_apply, 2, [file: 'lib/task/supervised.ex', line: 85]},\n    {Task.Supervised, :reply, 5, [file: 'lib/task/supervised.ex', line: 36]},\n    {:proc_lib, :init_p_do_apply, 3, [file: 'proc_lib.erl', line: 247]}]}}\n# ⇑⇑⇑⇑ one or more of ⇑⇑⇑⇑\n★: 42\n{:ok, 42}\n```","ref":"task_management.html#callbacks"},{"type":"extras","title":"Allowed options - Task Management","doc":"- `attempts` integer, an amount of attempts before giving up, `0` for forever; default: `:infinity`\n- `delay` the delay between attempts, default: `:none`;\n- `raise`: when `true`, will raise after all attempts were unsuccessful, or return `{:error, outcome}` tuple otherwise, default: `false`;\n- `accept_not_ok`: when `true`, any result save for `{:error, _}` will be accepted as correct, otherwise only `{:ok, _}` is treated as correct, default: `true`;\n- `on_success`: callback when the task returned a result, default: `nil`;\n- `on_retry`: callback when the task failed and scheduled for retry, default: `:debug`;\n- `on_fail`: callback when the task completely failed, default: `:warn`.\n\nfor `:attempts` and `:delay` keys one might specify the following values:\n\n- `integer` → amount to be used as is (milliseconds for `delay`, number for attempts);\n- `float` → amount of thousands, rounded (seconds for `delay`, thousands for attempts);\n- `:none` → `0`;\n- `:tiny` → `10`;\n- `:medium` → `100`;\n- `:infinity` → `-1`, `:attempts` only.","ref":"task_management.html#allowed-options"},{"type":"extras","title":"Task spawning - Task Management","doc":"```elixir\nTarearbol.run_in fn -> IO.puts(42) end, 1_000 # 1 sec\nTarearbol.spawn fn -> IO.puts(42) end # immediately\n```","ref":"task_management.html#task-spawning"},{"type":"extras","title":"Task draining - Task Management","doc":"```elixir\nTarearbol.run_in fn -> IO.inspect(42) end, 1_000 # 1 sec\nTarearbol.run_in fn -> IO.inspect(:foo) end, 1_000 # 1 sec\nTarearbol.drain\n42       # immediately, from `IO.inspect`\n:foo     # immediately, from `IO.inspect`\n[ok: 42, ok: :foo] # immediately, the returned value(s)\n```","ref":"task_management.html#task-draining"},{"type":"extras","title":"Dynamic Workers Management","doc":"# Dynamic Workers Management","ref":"dynamic_workers_management.html"},{"type":"extras","title":"Dynamic Management Examples - Dynamic Workers Management","doc":"There are some ready-to-use snippets [collected here](./dynamic_management_examples.md).","ref":"dynamic_workers_management.html#dynamic-management-examples"},{"type":"extras","title":"Architecture - Dynamic Workers Management","doc":"![DynamicManager Architecture](./assets/dynamic_manager.png)\n\n`DynamicSupervisor` is a very handy tool to manage many different task-like processes.\nIt is used mostly when one does not know in advance how many processes are to be run,\nand the processes happen to appear based on some external conditions. A good example\nwould be a web scraper; we might spawn a process per every new page we need to crawl\nand `DynamicSupervisor` would take care about restarting those failed.\n\nIt might become slightly cumbersome for the projects requiring more-or-less same\nbehaviour but when the supervised processes should perform some periodical job,\nmaybe exit upon job outcome and be accessible by some `id` while they are running.\n\nFor the contrived example, imagine the main dashboard that displays the state\nof all the computers, connected to the intranet. There are guests with notebooks,\nconstantly connecting and disconneting. Let’s say we want to maintain a list of\nprocesses spying on our guests (please do not use this library that way, though.)\n\nWe need to be able to query the processes for current state / stats by computer ID,\nsay, MAC-address. The first wild guess would be to convert MACs to alphanumeric\nand then to atoms to give names to our workers. And then call them by name.\nThis approach is not only naïve but also very dangerous. That way we’d DOS our\n_ErlangVM_ with atoms. So yeah, the dictionary `MAC → PID` is to be stored\nsomewhere else. So we already need to have a supervisor, managing the state _and_\n`DynamicSupervisor`, which in turn manages workers. Also, upon startup there is\na warming period during which we probably do not want to show anything since the\ndata might be inaccurate. This warming stage should be probably done from inside\n`handle_continue/2` from some another process, and this process cannot be the state\none due to restart strategy `:rest_for_one`, which is required to restart workers\nwhen the state has crashed.\n\nAll the problems above are very similar for this kind of task. So, welcome\n`DynamicSupervisor` which solves all the problems above automagically, leaving\nthe consumer with a pure business logic implementation. The only needed thing\nto make it all up and running would be to implement `Tarearbol.DynamicManager`\nbehaviour. It consists of four functions.\n\n###  `children_specs/0`\n\n```elixir\n@callback children_specs :: %{required(binary()) => Enum.t()}\n```\n\nReturn value should be a map of `id → settings` where `id` is the unique\nidentifier of the process and possible settings values are described in the\ndocumentation.\n\n### `perform/2`\n\n```elixir\n@callback perform(id :: binary(), payload :: term()) :: any()\n```\n\nThe implementation of the worker. This function will be called with\nthe child `id` as first argument and the `payload` option to child spec\nas second argument.\n\nIf it returns `:halt`, the process is considered done his job. Any other outcome\nwill be treated as as a result and stored in the `State`.\n\n### `handle_state_change/1`\n\n```elixir\n@callback handle_state_change(state :: :down | :up | :starting | :unknown) :: :ok | :restart\n```\n\nThis callback will be called on state changes, like `:starting` while the\ninitial state is not yet fully loaded and `:started` upon readyness.\n\n### `handle_timeout/1`\n\n```elixir\n@callback handle_timeout(state :: map()) :: any()\n```\n\nThis callback will be called if the worker cannot process in a reasonable time.","ref":"dynamic_workers_management.html#architecture"},{"type":"extras","title":"Dynamic Management Examples","doc":"# Dynamic Management Examples","ref":"dynamic_management_examples.html"},{"type":"extras","title":"Schedule Work - Dynamic Management Examples","doc":"```elixir\ndefmodule Counter do\n  alias Tarearbol.DynamicManager\n  use DynamicManager\n\n  @impl DynamicManager\n  def children_specs do\n    now = DateTime.utc_now()\n    %{\n      second: [payload: now, timeout: 1_000],\n      minute: [payload: now, timeout: 60_000]\n    }\n  end\n\n  @impl DynamicManager\n  def perform(:second, payload) do\n    now = DateTime.utc_now()\n    IO.inspect({payload, now}, label: \"sec\")\n    {:ok, now} # do not replace a payload\n  end\n\n  def perform(:minute, payload) do\n    now = DateTime.utc_now()\n    IO.inspect({payload, now}, label: \"min\")\n    {:replace, now} # replace a payload\n  end\nend\n```\n\nOnce started with `Counter.start_link()`, it’ll output:\n\n```\nsec: {~U[2021-03-10 07:58:55.874591Z], ~U[2021-03-10 07:58:56.876633Z]}\n...\nsec: {~U[2021-03-10 07:58:55.874591Z], ~U[2021-03-10 07:59:54.938543Z]}\nmin: {~U[2021-03-10 07:58:55.874591Z], ~U[2021-03-10 07:59:55.876592Z]}\nsec: {~U[2021-03-10 07:58:55.874591Z], ~U[2021-03-10 07:59:55.939661Z]}\n...\nsec: {~U[2021-03-10 07:58:55.874591Z], ~U[2021-03-10 08:00:54.998537Z]}\nmin: {~U[2021-03-10 07:59:55.876592Z], ~U[2021-03-10 08:00:55.877658Z]}\n...\n```\n\nNote, that the payload of the `:second` one does not change, while the payload of `minute` gets updated to the latest value reported.","ref":"dynamic_management_examples.html#schedule-work"},{"type":"extras","title":"Workers Pool - Dynamic Management Examples","doc":"One might create a workers pool with a help of `Tarearbol.DynamicManager`. Usually `perform/2` is suppressed in workers with `timeout: 0` and only `Genserver.cast/2` and `Genserver.call/2` are handled with helpers `Tarearbol.Pool.defsynch/2` and `Tarearbol.Pool.defasynch/2`.\n\nWritten as a regular functions, they are wrapped during compilation time to be dispatched to the free instance of the pool behind. Inside the body of these functions, the following magic macros become available:\n\n- **`id!`** returning the `id` of the worker invoked\n- **`payload!`** returning the `payload` of the worker invoked\n- **`state!`** returning the `state` of the worker invoked as a tuple `{id, payload}`\n\n```elixir\ndefmodule Pool do\n  use Tarearbol.Pool, init: &Pool.initializer/0, pool_size: 2\n\n  def initializer, do: 0\n\n  defsynch synch(),\n    do: {:ok, payload!()}\n\n  defsynch synch(n),\n    do: {:ok, payload!() + n}\n\n  defasynch asynch(n),\n    do: {:replace, payload!() + n}\nend\n```\n\nNow one might call `Pool.synch/1` to perform a synchronized pooled state request, as well as `Pool.asynch/1` to asynchronously update the state. Note, that in this contrived example, the state of the _first free worker_ is going to be updated.\n\n```elixir\nPool.start_link()\n\nEnum.map(1..3, &Pool.synch(&1))\n#⇒ [ok: 1, ok: 2, ok: 3]\nPool.synch()\n#⇒ 0 \n```\n\nBut for `asynch/1` function that needs some time to finish _and_ updates the state, it’s different.\n\n```elixir\nPool.start_link()\n\nEnum.map(1..3, &Pool.asynch(&1))\n#⇒ [:ok, :ok, :ok]\nEnum.reduce(Pool.state().children, 0, & &2 + elem(&1, 1).value)\n#⇒ 6\n```\n\nNote, that there is no guarantee what worker would handle each call.","ref":"dynamic_management_examples.html#workers-pool"},{"type":"extras","title":"Multiple Stateful Processes - Dynamic Management Examples","doc":"Consider we are building an online shop having buckets for each customer. Then we might back up the customer session with this kind of `DynamicManager`.\n\n```elixir\ndefmodule Bucket do\n  alias Tarearbol.DynamicManager\n  use DynamicManager\n\n  def new(customer) do\n    put(customer, payload: %{}, timeout: 10_000)\n  end\n\n  def do_smth(customer) do\n    IO.inspect(\"Hey, #{customer}, we have a discount\")\n  end\n\n  @impl DynamicManager\n  def children_specs, do: %{}\n\n  @impl DynamicManager\n  def perform(id, payload) do\n    if map_size(payload) > 0, do: do_smth(id)\n    {:ok, DateTime.utc_now()}\n  end\n\n  @impl DynamicManager\n  def call(:<, _from, {_id, payload}),\n    do: {:ok, payload}\n\n  @impl DynamicManager\n  def cast({:+, item}, {_id, payload}),\n    do: {:replace, Map.update(payload, item, 1, & &1+1)}\nend\n```\n\nThen we can play with it.\n\n```elixir\niex|1▸ Bucket.start_link()\n{:ok, #PID<0.325.0>}\niex|2▸ Bucket.new \"Aleksei\"  \n:ok\niex|3▸ Bucket.synch_call \"Aleksei\", {:+, :tomato}\niex|4▸ Bucket.asynch_call \"Aleksei\", {:+, :tomato}\niex|5▸ Bucket.asynch_call \"Aleksei\", {:+, :cucumber}\n# the below is printed from `perform/2`\n\"Hey, Aleksei, we have a discount\"                   \niex|6▸ Bucket.synch_call \"Aleksei\", :<\n{:ok, %{cucumber: 1, tomato: 2}}\n```","ref":"dynamic_management_examples.html#multiple-stateful-processes"},{"type":"extras","title":"Cron Management","doc":"# Cron Management","ref":"cron_management.html"},{"type":"extras","title":"Intro - Cron Management","doc":"`Tarearbol` provides a simple but robust support of cron-like jobs scheduling. The approach is to spawn a process for each job to be completed under `Tarearbol.DynamicManager` supervision. `Tarearbol` application _does not_ start the `Tarearbol.Scheduler` automatically, the target application must include it into it’s supervision tree explicitly.\n\nJobs description are loaded upon application start from one of the following locations:\n\n- `project` section of the `Mix.Project`, key `:tarearbol`, subkey `:jobs`\n- `config` entry in `Config` script, key `:tarearbol`, subkey `:jobs`\n- `.tarearbol.exs` file in the root folder of the target project (the path to the file might be configured in `project` section  of the `Mix.Project`, key `:tarearbol`, subkey `:jobs_file`)\n\n`Tarearbol.Scheduler` understands `cron` format. It exports three handy methods from `Tarearbol.Crontab` module, `Tarearbol.Crontab.next/3`, `Tarearbol.Crontab.next_as_stream/3` and `Tarearbol.Crontab.next_as_list/3`. The first might be used to get the timestamp of the _next_ event by cron record, the second to greedy evaluate all the next events for the year and the last one to return a `Stream` lazily evaluating next events.\n\n---","ref":"cron_management.html#intro"},{"type":"extras","title":"Job Management - Cron Management","doc":"`Tarearbol.Scheduler` was created mostly for the single-node application, but it might be easily extended to manage jobs on several nodes.\n\nUpon start, it loads jobs schedules and spawns processes for each, managed by `Tarearbol.DynamicManager`. Each job must return on of the following three outcomes.\n\n- `{:ok, any()}` to normally return the result to be stored as last job execution result in the state of `Manager` and reschedule the job to the next event\n- `:halt` to prevent further job executions and remove it from the list of scheduled jobs\n- `{{:reschedule, binary()}, any()}` to reschedule the job with new cron record, given as the second parameter of the first tuple.\n\n---","ref":"cron_management.html#job-management"},{"type":"extras","title":"Extra Sugar - Cron Management","doc":"`Tarearbol.Calendar` exposes two functions stolen from _Rails_:\n\n- `Tarearbol.Calendar.beginning_of/2` returns the beginning of the period including the timestamp given as an argument\n- `Tarearbol.Calendar.end_of/2` returns the end of the period including the timestamp.","ref":"cron_management.html#extra-sugar"}],"content_type":"text/markdown"}